module Dynamics where

open import Prelude

module MonoDyn where
  data U : Set where
    NAT   : U
    PAIR  : U -> U -> U
    _=>_  : U -> U -> U

  el : U -> Set
  el NAT          = Nat
  el (PAIR u1 u2)  = Pair (el u1) (el u2)
  el (u1 => u2)    = el u1 -> el u2

  data Dynamic : Set where
    Dyn : (u : U) -> el u -> Dynamic
  decEqU : (u1 u2 : U) -> Either (u1 == u2) (u1 <> u2)

  pairLeft : {x x' y y' : U} -> PAIR x y == PAIR x' y' -> x == x'
  pairLeft Refl = Refl

  pairRight : {x x' y y' : U} -> PAIR x y == PAIR x' y' -> y == y'
  pairRight Refl = Refl

  funLeft : {x x' y y' : U} -> x => y == x' => y' -> x == x'
  funLeft Refl = Refl

  funRight : {x x' y y' : U} -> x => y == x' => y' -> y == y'
  funRight Refl = Refl

  decEqU NAT NAT = Inl Refl
  decEqU NAT (PAIR y y') = Inr \()
  decEqU NAT (y => y') = Inr \()
  decEqU (PAIR y y') NAT = Inr \()
  decEqU (PAIR y y') (PAIR y0 y1) with decEqU y y0 | decEqU y' y1
  decEqU (PAIR .y0 .y1) (PAIR y0 y1) | Inl Refl | Inl Refl = Inl Refl
  decEqU (PAIR y y') (PAIR y0 y1) | Inl y2 | Inr y3 = Inr (\p -> magic (y3 (pairRight p)))
  decEqU (PAIR y y') (PAIR y0 y1) | Inr y2 | p2 = Inr (\p -> magic (y2 (pairLeft p)))
  decEqU (PAIR y y') (y0 => y1) = Inr \()
  decEqU (y => y') NAT = Inr \()
  decEqU (y => y') (PAIR y0 y1) = Inr \()
  decEqU (y => y') (y0 => y1) with decEqU y y0 | decEqU y' y1
  decEqU (.y0 => .y1) (y0 => y1) | Inl Refl | Inl Refl = Inl Refl
  decEqU (y => y') (y0 => y1) | Inl y2 | Inr y3 = Inr (\p -> magic (y3 (funRight p)))
  decEqU (y => y') (y0 => y1) | Inr y2 | p2 = Inr (\p -> magic (y2 (funLeft p)))

  cast : (u1 : U) -> Dynamic -> Maybe (el u1)
  cast u1 (Dyn u2 x) with decEqU u1 u2
  cast u1 (Dyn .(u1) x)  | Inl Refl   = Just x
  cast u1 (Dyn u2 x)         | _          = Nothing

module PolyDyn where

  data Fin : Nat -> Set where
    Fz  : forall { n} -> Fin (Succ n)
    Fs  : forall { n} -> Fin n -> Fin (Succ n)

  fsInj : forall {n} {i j : Fin n} -> Fs i == Fs j -> i == j
  fsInj Refl = Refl

  decEqFin : {n : Nat} -> (i j : Fin n) -> Either (i == j) (i <> j)
  decEqFin Fz Fz = Inl Refl
  decEqFin Fz (Fs y) = Inr \()
  decEqFin (Fs y) Fz = Inr \()
  decEqFin (Fs y) (Fs y') with decEqFin y y' 
  decEqFin (Fs .y') (Fs y') | Inl Refl = Inl Refl
  ... | Inr q = Inr (\p -> magic (q (fsInj p)))

  data U (n : Nat) : Set where
    NAT   :  U n
    PAIR  :  U n -> U n -> U n
    _=>_  :  U n -> U n -> U n
    VAR   :  Fin n -> U n

  pairLeft : {n : Nat} {x x' y y' : U n} -> PAIR x y == PAIR x' y' -> x == x'
  pairLeft Refl = Refl

  pairRight : {n : Nat} {x x' y y' : U n} -> PAIR x y == PAIR x' y' -> y == y'
  pairRight Refl = Refl

  funLeft : {n : Nat} {x x' y y' : U n} -> x => y == x' => y' -> x == x'
  funLeft Refl = Refl

  funRight : {n : Nat} {x x' y y' : U n} -> x => y == x' => y' -> y == y'
  funRight Refl = Refl

  varInj : {n : Nat} {i j : Fin n} -> VAR i == VAR j -> i == j
  varInj Refl = Refl

  data V : Set where
    FORALL : {n : Nat} -> U n -> V

  forallInj : {n : Nat} {u u' : U n} -> FORALL u == FORALL u' -> u == u' 
  forallInj Refl = Refl

  forallNInj : {n n' : Nat} {u : U n} {u' : U n'} -> FORALL u == FORALL u' -> n == n'
  forallNInj Refl = Refl

  idType : V
  idType = FORALL {Succ Zero} (VAR Fz => VAR Fz) 

  data Env : Nat -> Set where
    Nil   : Env Zero
    Cons  : forall {n : Nat} -> U Zero -> Env n -> Env (Succ n)

  findInEnv : forall {n} -> Fin n -> Env n -> U Zero
  findInEnv Fz      (Cons u _)    = u
  findInEnv (Fs i)  (Cons _ env)  = findInEnv i env

  close : forall { n} -> U n -> Env n -> U Zero
  close NAT           _    = NAT
  close (PAIR u1 u2)  env  = PAIR (close u1 env) (close u2 env)
  close (u1 => u2)    env  = close u1 env => close u2 env
  close (VAR i)       env  = findInEnv i env

  elClosed : U Zero -> Set
  elClosed NAT           =  Nat
  elClosed (PAIR u1 u2)  =  Pair (elClosed u1) (elClosed u2)
  elClosed (u1 => u2)    =  elClosed u1 -> elClosed u2
  elClosed (VAR ())

  elU : forall {n} -> U n -> Env n -> Set
  elU u env = elClosed (close u env)

  elV : V -> Set
  elV (FORALL {n} u) = {env : Env n} -> elU u env

  decEqU : forall {n} -> (u1 u2 : U n) -> Either (u1 == u2) (u1 <> u2)
  decEqU NAT NAT = Inl Refl
  decEqU NAT (PAIR y y') = Inr \()
  decEqU NAT (y => y') = Inr \()
  decEqU NAT (VAR i) = Inr \()
  decEqU (PAIR y y') NAT = Inr \()
  decEqU (PAIR y y') (PAIR y0 y1) with decEqU y y0 | decEqU y' y1
  decEqU (PAIR .y0 .y1) (PAIR y0 y1) | Inl Refl | Inl Refl = Inl Refl
  decEqU (PAIR y y') (PAIR y0 y1) | Inl y2 | Inr y3 = Inr (\p -> magic (y3 (pairRight p)))
  decEqU (PAIR y y') (PAIR y0 y1) | Inr y2 | p2 = Inr (\p -> magic (y2 (pairLeft p)))
  decEqU (PAIR y y') (y0 => y1) = Inr \()
  decEqU (PAIR y y') (VAR i) = Inr \()
  decEqU (y => y') NAT = Inr \()
  decEqU (y => y') (PAIR y0 y1) = Inr \()
  decEqU (y => y') (y0 => y1) with decEqU y y0 | decEqU y' y1
  decEqU (.y0 => .y1) (y0 => y1) | Inl Refl | Inl Refl = Inl Refl
  decEqU (y => y') (y0 => y1) | Inl y2 | Inr y3 = Inr (\p -> magic (y3 (funRight p)))
  decEqU (y => y') (y0 => y1) | Inr y2 | p2 = Inr (\p -> magic (y2 (funLeft p)))
  decEqU (y => y') (VAR i) = Inr \()
  decEqU (VAR i) NAT = Inr \()
  decEqU (VAR i) (PAIR y y') = Inr \()
  decEqU (VAR i) (y => y') = Inr \()
  decEqU (VAR i) (VAR j) with decEqFin i j
  decEqU (VAR .j) (VAR j) | Inl Refl = Inl Refl
  decEqU (VAR i) (VAR j) | Inr neq = Inr (\eq -> neq (varInj eq))

  data Dynamic : Set where
    Dyn : (v : V) -> elV v -> Dynamic

  idDyn : Dynamic
  idDyn = Dyn idType (\ x -> x)

  decEqV : (v1 : V) -> (v2 : V) -> Either (v1 == v2) (v1 <> v2)
  decEqV (FORALL {n} u) (FORALL {m} u') with decEqNat n m
  decEqV (FORALL u) (FORALL u') | Inl Refl with decEqU u u'
  ... | Inl p = Inl (cong FORALL p)
  ... | Inr q = Inr (\eq -> q (forallInj eq))
  decEqV (FORALL u) (FORALL u') | Inr q = Inr (\eq -> q (forallNInj eq))

  module CastI where
    cast : (v1 : V) -> Dynamic -> Maybe (elV v1)
    cast v1 (Dyn v2 x) with decEqV v1 v2
    cast v1 (Dyn .(v1) x)  | Inl Refl   = Just x
    ...                        | _          = Nothing

    success : cast idType idDyn == Just (\x -> x)
    success = Refl

    fail : cast (FORALL {Zero} (NAT => NAT)) idDyn == Nothing
    fail = Refl

  module Subst where

    data Subst (m : Nat) : Nat -> Set where
      Nil  : Subst m Zero
      Cons : forall {n} -> Maybe (U m) -> Subst m n -> Subst m (Succ n)

    consInjHead : forall {n m : Nat} {mu mu' : Maybe (U m)} {subst subst' : Subst m n} -> 
      _==_ {zero} {Subst m (Succ n)} (Cons mu subst) (Cons mu' subst') -> mu == mu'
    consInjHead Refl = Refl

    consInjTail : forall {n m : Nat} {mu mu' : Maybe (U m)} {subst subst' : Subst m n} -> 
      _==_ {zero} {Subst m (Succ n)} (Cons mu subst) (Cons mu' subst') -> subst == subst'
    consInjTail Refl = Refl

    empty : forall { n} -> {m : Nat} -> Subst m n
    empty {Zero}    = Nil
    empty {Succ n}  = Cons Nothing empty

    findInSubst : forall { n m} -> Fin n -> Subst m n -> Maybe (U m)
    findInSubst Fz      (Cons x _)      = x
    findInSubst (Fs i)  (Cons _ subst)  = findInSubst i subst

    apply : forall { n} -> Subst n n -> U n -> U n
    apply _      NAT           = NAT
    apply subst  (PAIR u1 u2)  = PAIR (apply subst u1) (apply subst u2)
    apply subst  (u1 => u2)    = apply subst u1 => apply subst u2
    apply subst  (VAR i)  with findInSubst i subst
    ...                   | Just u   = u
    ...                   | Nothing  = VAR i

    extend :  forall { n m} -> Fin n -> U m -> Subst m n -> Maybe (Subst m n)
    extend Fz      u  (Cons Nothing subst) = Just (Cons (Just u) subst)
    extend Fz      u  (Cons (Just u') _) with decEqU u u'
    extend Fz      u  (Cons (Just .(u)) subst)  | Inl Refl  = Just (Cons (Just u) subst) 
    extend Fz      u  (Cons (Just u') _)            | Inr _     = Nothing
    extend (Fs i)  u  (Cons mu' subst) with extend i u subst
    ... | Just subst'  = Just (Cons mu' subst')
    ... | _            = Nothing

    lookupExtend : forall {n m : Nat} -> (u : U m) -> (i : Fin n) -> (subst subst' : Subst m n) ->
      extend i u subst == Just subst' -> findInSubst i subst' == Just u
    lookupExtend u Fz (Cons Nothing _) (Cons u' _) eq = consInjHead (justInj (sym eq))
    lookupExtend u Fz (Cons (Just u') _) (Cons _ _) eq with decEqU u u'
    lookupExtend u Fz (Cons (Just .u) _) (Cons _ _) eq | Inl Refl = consInjHead (justInj (sym eq)) 
    lookupExtend _ Fz (Cons (Just _) _) (Cons _ _) () | Inr _
    lookupExtend u (Fs i) (Cons _ subst) (Cons _ subst') eq with inspect (extend i u subst)
    ... | Just _ by eq' rewrite eq' | consInjTail (justInj eq) = lookupExtend u i subst subst' eq' 
    ... | Nothing by eq' rewrite eq' = magic (justNothing eq)

    mutual
      check : forall {n : Nat} -> U n -> U n -> Maybe (Subst n n)
      check u1 u2 = checkAcc u1 u2 empty
      checkAcc : forall {n} -> U n -> U n -> Subst n n -> Maybe (Subst n n)
      checkAcc NAT           NAT             subst  = Just subst
      checkAcc (PAIR u1 u2)  (PAIR w1 w2)    subst with checkAcc u1 w1 subst
      ... | Just subst'  = checkAcc u2 w2 subst'
      ... | _            = Nothing
      checkAcc (u1 => u2)    (w1 => w2)      subst with checkAcc u1 w1 subst
      ... | Just subst'  = checkAcc u2 w2 subst'
      ... | _            = Nothing
      checkAcc u             (VAR i)         subst   = extend i u subst
      checkAcc _             _               _      = Nothing

    step : forall {n : Nat} -> (u : U n) -> (i : Fin n) -> (subst : Subst n n) ->
      checkAcc u (VAR i) subst == extend i u subst
    step NAT i subst = Refl
    step (PAIR u2 u3) i subst = Refl
    step (u2 => u3) i subst = Refl
    step (VAR j) i subst = Refl

    checkAccCorrect : forall {n} -> (u1 u2 : U n) -> (subst subst' : Subst n n) -> 
      checkAcc u1 u2 subst == Just subst' -> 
      apply subst' u2 == u1
    stability : forall { n} -> (u1 u2 w1 w2 : U n) (subst subst' subst'' : Subst n n) -> 
      checkAcc u1 u2 subst == Just subst' -> 
      checkAcc w1 w2 subst' == Just subst'' -> 
      apply subst'' u2 == apply subst' u2

    findStability : forall {n : Nat} -> (u u1 u2 : U n) -> (i : Fin n) -> (subst subst' : Subst n n) ->
      findInSubst i subst == Just u -> checkAcc u1 u2 subst == Just subst' -> findInSubst i subst' == Just u
    findStability u NAT NAT i subst subst' eq eq' rewrite justInj eq' = eq
    findStability u NAT (PAIR u1 u2) i subst subst' eq ()
    findStability u NAT (u1 => u2) i subst subst' eq ()
    findStability u (PAIR u1 u2) NAT i subst subst' eq ()
    findStability u (PAIR u1 u2) (PAIR u3 u4) i subst subst' eq eq' with inspect (checkAcc u1 u3 subst)
    ... | Just subst'' by eq'' 
      rewrite eq'' 
      | findStability u u2 u4 i subst'' subst' (findStability u u1 u3 i subst subst'' eq eq'') eq' = Refl
    ... | Nothing by eq'' rewrite eq'' = magic (justNothing eq')
    findStability u (PAIR u1 u2) (u3 => u4) i subst subst' eq ()
    findStability u (u1 => u2) NAT i subst subst' eq ()
    findStability u (u1 => u2) (PAIR u3 u4) i subst subst' eq ()
    findStability u (u1 => u2) (u3 => u4) i subst subst' eq eq' with inspect (checkAcc u1 u3 subst)
    ... | Just subst'' by eq'' 
      rewrite eq'' 
      | findStability u u2 u4 i subst'' subst' (findStability u u1 u3 i subst subst'' eq eq'') eq' = Refl
    ... | Nothing by eq'' rewrite eq'' = magic (justNothing eq')
    findStability u (VAR y) NAT i subst subst' eq ()
    findStability u (VAR y) (PAIR u1 u2) i subst subst' eq ()
    findStability u (VAR y) (u1 => u2) i subst subst' eq ()
    findStability u u1 (VAR i) j subst subst' eq eq' 
      rewrite step u1 i subst 
      with decEqFin i j | decEqU u u1 
    findStability u .u (VAR .j) j subst subst' eq eq' | Inl Refl | Inl Refl 
      = lookupExtend u j subst subst' eq'
    findStability u u1 (VAR .i) i subst subst' eq eq' | Inl Refl | Inr q 
      = magic (q (updateEq u u1 i subst subst' eq eq'))
      where
      updateEq : forall {n m : Nat} -> (u u' : U m) -> (i : Fin n) -> (subst subst' : Subst m n) ->
        findInSubst i subst == Just u -> extend i u' subst == Just subst' -> u == u'
      updateEq _ _ Fz (Cons Nothing _) (Cons _ _) () _
      updateEq _ u' Fz (Cons (Just u'') _) (Cons _ _) eq _ with decEqU u' u''
      updateEq _ u' Fz (Cons (Just .u') _) (Cons _ _) eq _ | Inl Refl rewrite justInj (sym eq) = Refl
      updateEq _ _ Fz (Cons (Just _) _) (Cons _ _) _ () | Inr _
      updateEq u u' (Fs i) (Cons _ subst)  (Cons _ _) eq eq' with inspect (extend i u' subst)
      ... | Just subst'' by eq'' rewrite eq'' | updateEq u u' i subst subst'' eq eq'' = Refl
      ... | Nothing by eq'' rewrite eq'' = magic (justNothing eq')
    findStability u u1 (VAR i) j subst subst' eq eq' | Inr neq | p2 = 
      updateNeq u u1 j i subst subst' neq eq eq'
      where
      updateNeq : forall {n m : Nat} -> (u u' : U m) -> (i j : Fin n) -> (subst subst' : Subst m n) -> 
         j <> i -> findInSubst i subst == Just u -> extend j u' subst == Just subst' -> 
         findInSubst i subst' == Just u
      updateNeq _ _ Fz Fz (Cons Nothing _) _ neq _ _ = magic (neq Refl)
      updateNeq _ _ (Fs _) Fz (Cons Nothing _) (Cons Nothing _) _ _ ()
      updateNeq _ _ (Fs _) Fz (Cons Nothing _) (Cons (Just _) _) _ eq' eq'' 
        rewrite consInjTail (justInj eq'') | eq' = Refl
      updateNeq _ u' _ Fz (Cons (Just u'') _) _ _ _ eq'' with decEqU u' u''
      updateNeq _ u' _ Fz (Cons (Just .u') _) _ _ eq' eq'' | Inl Refl rewrite justInj eq'' | eq' = Refl
      updateNeq _ _ _ Fz (Cons (Just _) _) _ _ _ () | Inr _
      updateNeq _ u' Fz (Fs j) (Cons _ subst) (Cons _ _) _ eq' eq'' with extend j u' subst
      ... | Just _ rewrite eq' | consInjHead (justInj (sym eq'')) = Refl
      updateNeq _ _ Fz (Fs _) (Cons _ _) (Cons _ _) _ _ () | Nothing
      updateNeq u u' (Fs i) (Fs j) (Cons _ subst) (Cons _ subst') neq eq' eq'' with inspect (extend j u' subst)
      ... | Just _ by eq''' rewrite eq''' | consInjTail (justInj eq'') 
        = updateNeq u u' i j subst subst' (\eq -> neq (cong Fs eq)) eq' eq'''
      ... | Nothing by eq''' rewrite eq''' = magic (justNothing eq'')

    stability NAT NAT u1' u2' subst subst' subst'' eq eq' = Refl
    stability NAT (PAIR y y') u1' u2' subst subst' subst'' () eq'
    stability NAT (y => y') u1' u2' subst subst' subst'' () eq'
    stability (PAIR y y') NAT u1' u2' subst subst' subst'' () eq'
    stability (PAIR u1 u2) (PAIR u1' u2') u u' subst subst' subst'' eq eq' 
      with inspect (checkAcc u1 u1' subst)
    stability (PAIR u1 u2) (PAIR u1' u2') u u' subst subst' subst'' eq eq' | Nothing by eq'' 
      rewrite eq'' = magic (justNothing eq)
    stability (PAIR u1 u2) (PAIR u1' u2') u u' subst subst' subst'' eq eq' | Just subst''' by eq'' 
      rewrite eq'' | stability u2 u2' u u' subst''' subst' subst'' eq eq' 
      = cong (\u -> PAIR u (apply subst' u2')) (trans p (sym q))
      where
        lemma : checkAcc (PAIR u2 u) (PAIR u2' u') subst''' == Just subst''
        lemma with inspect (checkAcc u2 u2' subst''')
        ... | Just _ by _ rewrite eq | eq' = Refl
        ... | Nothing by eq''' = magic (justNothing (trans (sym eq''') eq))

        p : apply subst'' u1' == apply subst''' u1'
        p = stability u1 u1' (PAIR u2 u) (PAIR u2' u') subst subst''' subst'' eq'' lemma
      
        q : apply subst' u1' == apply subst''' u1'
        q = stability u1 u1' u2 u2' subst subst''' subst' eq'' eq
    stability (PAIR y y') (y0 => y1) u1' u2' subst subst' subst'' () eq'
    stability (y => y') NAT u1' u2' subst subst' subst'' () eq'
    stability (y => y') (PAIR y0 y1) u1' u2' subst subst' subst'' () eq'
    stability (u1 => u2) (u1' => u2') u u' subst subst' subst'' eq eq' 
      with inspect (checkAcc u1 u1' subst)
    stability (u1 => u2) (u1' => u2') u u' subst subst' subst'' eq eq' | Nothing by eq''
      rewrite eq'' = magic (justNothing eq)
    stability (u1 => u2) (u1' => u2') u u' subst subst' subst'' eq eq' | Just subst''' by eq'' 
      rewrite eq'' | stability u2 u2' u u' subst''' subst' subst'' eq eq'
      = cong (\u -> u => apply subst' u2') (trans p (sym q))
        where
        lemma : checkAcc (u2 => u) (u2' => u') subst''' == Just subst''
        lemma with inspect (checkAcc u2 u2' subst''')
        ... | Just _ by _ rewrite eq | eq' = Refl
        ... | Nothing by eq''' = magic (justNothing (trans (sym eq''') eq))

        p : apply subst'' u1' == apply subst''' u1'
        p = stability u1 u1' (u2 => u) (u2' => u') subst subst''' subst'' eq'' lemma
      
        q : apply subst' u1' == apply subst''' u1'
        q = stability u1 u1' u2 u2' subst subst''' subst' eq'' eq
    stability (VAR i) NAT u1' u2' subst subst' subst'' () eq'
    stability (VAR i) (PAIR y' y0) u1' u2' subst subst' subst'' () eq'
    stability (VAR i) (y' => y0) u1' u2' subst subst' subst'' () eq'
    stability u1 (VAR i) u1' u2' subst subst' subst'' eq eq' 
      rewrite step u1 i subst 
      with lookupExtend u1 i subst subst' eq 
    ... | eq'' rewrite eq'' | findStability u1 u1' u2' i subst' subst'' eq'' eq' = Refl

    checkAccCorrect NAT NAT subst subst' eq = Refl
    checkAccCorrect NAT (PAIR u1 u2) subst subst' ()
    checkAccCorrect NAT (u1 => u2) subst subst' ()
    checkAccCorrect (PAIR u1 u2) NAT subst subst' ()
    checkAccCorrect (PAIR u1 u2) (PAIR u3 u4) subst subst' eq 
      with inspect (checkAcc u1 u3 subst)
    checkAccCorrect (PAIR u1 u2) (PAIR u3 u4) subst subst' eq | Nothing by eq'
      rewrite eq' = magic (justNothing eq)
    checkAccCorrect (PAIR u1 u2) (PAIR u3 u4) subst subst' eq | Just subst'' by eq' 
      rewrite eq' 
      | checkAccCorrect u2 u4 subst'' subst' eq 
      | stability u1 u3 u2 u4 subst subst'' subst' eq' eq 
      | checkAccCorrect u1 u3 subst subst'' eq' = Refl
    checkAccCorrect (PAIR u1 u2) (u3 => u4) subst subst' ()
    checkAccCorrect (u1 => u2) NAT subst subst' ()
    checkAccCorrect (u1 => u2) (PAIR u3 u4) subst subst' ()
    checkAccCorrect (u1 => u2) (u3 => u4) subst subst' eq 
      with inspect (checkAcc u1 u3 subst)
    checkAccCorrect (u1 => u2) (u3 => u4) subst subst' eq | Nothing by eq' 
      rewrite eq' = magic (justNothing eq)
    checkAccCorrect (u1 => u2) (u3 => u4) subst subst' eq | Just subst'' by eq'
      rewrite eq' 
      | checkAccCorrect u2 u4 subst'' subst' eq 
      | stability u1 u3 u2 u4 subst subst'' subst' eq' eq 
      | checkAccCorrect u1 u3 subst subst'' eq' = Refl
    checkAccCorrect (VAR y) NAT subst subst' ()
    checkAccCorrect (VAR y) (PAIR u1 u2) subst subst' ()
    checkAccCorrect (VAR y) (u1 => u2) subst subst' ()
    checkAccCorrect u (VAR i) subst subst' eq 
      rewrite step u i subst
      | lookupExtend u i subst subst' eq = Refl

    instantiate : forall {n} -> (u : U n) -> (subst : Subst n n) -> 
      elV (FORALL {n} u) -> elV (FORALL {n} (apply subst u))

    instantiateEnv : forall {n} -> Subst n n -> Env n -> Env n

    instantiateEnvCorrect : forall { n} -> (u : U n) (subst : Subst n n) (env : Env n) ->
      elU u (instantiateEnv subst env) == elU (apply subst u) env

    instantiate u subst f {env} = 
       coerce (instantiateEnvCorrect u subst env) (f {instantiateEnv subst env})
      where  
         coerce : forall { a b : Set} -> a == b -> a -> b
         coerce Refl x = x

    build : forall { n m} -> Subst m n -> Env n -> Env m -> Env n
    build Nil                    _             _     = Nil
    build (Cons Nothing subst)   (Cons u env)  env'  = Cons u (build subst env env')
    build (Cons (Just u) subst)  (Cons _ env)  env'  =
      Cons (close u env') (build subst env env')

    instantiateEnv subst env = build subst env env

    notInSubst : forall {n m : Nat} -> (i : Fin n) -> (subst : Subst m n) -> (env : Env n) (env' : Env m) ->
      findInSubst i subst == Nothing -> findInEnv i env == findInEnv i (build subst env env')
    notInSubst Fz (Cons Nothing y') (Cons y0 y1) env' eq = Refl
    notInSubst Fz (Cons (Just y) y') (Cons y0 y1) env' ()
    notInSubst (Fs i) (Cons Nothing subst) (Cons y env) env' eq = notInSubst i subst env env' eq
    notInSubst (Fs i) (Cons (Just y) subst) (Cons y' env) env' eq = notInSubst i subst env env' eq

    inSubst : forall {n m : Nat} -> (i : Fin n) -> (u : U m) -> (subst : Subst m n) -> 
      (env : Env n) -> (env' : Env m) ->
      findInSubst i subst == Just u -> elU u env' == elU (VAR i) (build subst env env')
    inSubst Fz u (Cons Nothing _) env env' ()
    inSubst Fz u (Cons (Just u') subst) (Cons _ env) env' eq rewrite justInj eq = Refl
    inSubst (Fs i) u (Cons Nothing subst) (Cons _ env) env' eq = inSubst i u subst env env' eq
    inSubst (Fs i) u (Cons (Just _) subst) (Cons _ env) env' eq = inSubst i u subst env env' eq

    instantiateEnvCorrect NAT subst env = Refl
    instantiateEnvCorrect (PAIR u u') subst env 
      rewrite instantiateEnvCorrect u subst env
      | instantiateEnvCorrect u' subst env = Refl
    instantiateEnvCorrect (u => u') subst env 
      rewrite instantiateEnvCorrect u subst env
      | instantiateEnvCorrect u' subst env = Refl
    instantiateEnvCorrect (VAR i) subst env with inspect (findInSubst i subst)
    instantiateEnvCorrect (VAR i) subst env | Nothing by eq 
      rewrite eq 
      | notInSubst i subst env env eq = Refl
    instantiateEnvCorrect (VAR i) subst env | Just u by eq 
      rewrite eq 
      | inSubst i u subst env env eq = Refl

  module CastII where

    open Subst

    cast : (v1 : V) -> Dynamic -> Maybe (elV v1)
    cast (FORALL {n} u1) (Dyn (FORALL {m} u2) x2) with decEqNat n m
    cast (FORALL u1) (Dyn (FORALL u2) x) | Inr neq = Nothing
    cast (FORALL u1) (Dyn (FORALL u2) x) | Inl Refl with inspect (check u1 u2)
    cast (FORALL u1) (Dyn (FORALL u2) x) | Inl Refl | Nothing by eq = Nothing
    cast (FORALL u1) (Dyn (FORALL u2) x) | Inl Refl | Just subst by eq 
      = Just (coerce prf x')
        where
        x' : elV (FORALL (apply subst u2))
        x' = instantiate u2 subst x
        prf : apply subst u2 == u1
        prf = checkAccCorrect u1 u2 empty subst eq
        coerce : forall { n} -> { u1 u2 : U n} -> u1 == u2 -> elV (FORALL u1) -> elV (FORALL u2)
        coerce Refl f = f

    idNat : V
    idNat = FORALL {Zero} (NAT => NAT)

    test = cast idNat idDyn
    idNat1 : V
    idNat1 = FORALL {Succ Zero} (NAT => NAT)

module CastIII where

  open PolyDyn
  open Subst

  data Cmp : Nat -> Nat -> Set where
    EQ : forall {n} -> Cmp n n
    LT : forall {n} -> (k : Nat) -> Cmp n (k + n)
    GT : forall {n} -> (k : Nat) -> Cmp (k + n) n

  cmp : (n m : Nat) -> Cmp n m

  plusZero : (n : Nat) -> n + Zero == n

  plusSucc : (n m : Nat) -> n + Succ m == Succ (n + m)  

  coerce : forall {a : Level} {x y : Set a}-> x == y -> x -> y
  coerce Refl f = f

  plusZero Zero = Refl
  plusZero (Succ k) = cong Succ (plusZero k)

  plusSucc Zero m = Refl
  plusSucc (Succ k) m = cong Succ (plusSucc k m)

  cmp Zero Zero = EQ
  cmp Zero (Succ k) = 
    coerce (cong (\n -> Cmp Zero (Succ n)) (plusZero k)) (LT {Zero} (Succ k))
  cmp (Succ k) Zero = 
    coerce (cong (\n -> Cmp (Succ n) Zero) (plusZero k)) (GT {Zero} (Succ k))
  cmp (Succ k) (Succ k') with cmp k k'
  cmp (Succ .k') (Succ k') | EQ = EQ
  cmp (Succ k) (Succ .(k' + k)) | LT k' = 
    coerce (cong (\n -> Cmp (Succ k) n) (plusSucc k' k)) (LT {Succ k} k')
  cmp (Succ .(k + k')) (Succ k') | GT k = 
    coerce (cong (\n -> Cmp n (Succ k')) (plusSucc k k')) (GT {Succ k'} k)

  weaken : forall { n} -> U n -> U (Succ n)
  weaken NAT           = NAT
  weaken (PAIR u1 u2)  = PAIR (weaken u1) (weaken u2) 
  weaken (u1 => u2)    = weaken u1 => weaken u2
  weaken (VAR i)       = VAR (Fs i)

  weakenByK : forall {n} -> (k : Nat) -> U n -> U (k + n)
  weakenByK Zero      u = u
  weakenByK (Succ k)  u = weaken (weakenByK k u)

  strengthen : forall {n} -> (k : Nat) -> (u : U n) -> 
    elV (FORALL {k + n} (weakenByK k u)) -> elV (FORALL {n} u)

  pad : forall {n} -> Env n -> Env (Succ n)
  pad env = Cons NAT env

  padByK : forall {n : Nat} -> (k : Nat) -> Env n -> Env (k + n)
  padByK {n} Zero env = env
  padByK {n} (Succ k) env = pad (padByK k env)

  strengthenable : forall {n : Nat} -> (u : U n) -> (env : Env n) -> elU (weaken u) (pad env) == elU u env
  strengthenable NAT _ = Refl
  strengthenable (PAIR u u') env rewrite strengthenable u env | strengthenable u' env = Refl
  strengthenable (u => u') env rewrite strengthenable u env | strengthenable u' env = Refl
  strengthenable (VAR i) env = Refl

  strengthenableByK : forall {n : Nat} -> (k : Nat) -> (u : U n) -> (env : Env n) -> 
    elU (weakenByK k u) (padByK k env) == elU u env
  strengthenableByK {n} Zero _ _ = Refl
  strengthenableByK {n} (Succ k) u env
    rewrite strengthenable (weakenByK k u) (padByK k env) = strengthenableByK k u env

  strengthen {n} k u f {env} = strengthen' (f {padByK k env})
    where 
      strengthen' : elU (weakenByK k u) (padByK k env) -> elU u env
      strengthen' x rewrite strengthenableByK k u env = x

  cast : (v : V) -> Dynamic -> Maybe (elV v)
  cast (FORALL {n} _) (Dyn (FORALL {m} _) _) with cmp n m
  cast (FORALL {n} u1) (Dyn (FORALL {.n} u2) f) | EQ 
    with inspect (check u1 u2)
  ... | Just subst by eq =
          Just (\ {env} -> 
            let prf = checkAccCorrect u1 u2 empty subst eq in
            let correct = cong (\ u -> elClosed (close u env)) prf in
            let val = instantiate u2 subst f in
            coerce correct val)
  ... | _ = Nothing
  cast (FORALL u1) (Dyn (FORALL u2) f) | LT k 
    with inspect (check (weakenByK k u1) u2)
  ... | Just subst by eq = 
          Just (strengthen k u1 (\ {env} -> 
            let prf = checkAccCorrect (weakenByK k u1) u2 empty subst eq in
            let correct = cong (\ u -> elU u env) prf in
            let val = instantiate u2 subst f in
            coerce correct val))
  ... | _ = Nothing
  cast (FORALL u1) (Dyn (FORALL u2) f) | GT k = Nothing
